from rbtexceptions import RbtConfigError
from rbtenums import *
import os
from pwd import getpwuid
from os.path import exists, expanduser, dirname
from sys import stderr
from subprocess import check_call, check_output, CalledProcessError
from stat import S_IMODE
from re import match

EXIT_CODE_ERROR = 1
DEVNULL = open(os.devnull, 'wb')


def analyze_check(checks):
    """
    Анализирует результаты работы функций check_*, возвращая код статуса настройки сервиса:
    STATUS_OK -- Сервис настроен
    STATUS_ERROR -- Сервис настроен с ошибкой
    STATUS_NOT_SET -- Сервис не настроен
    :param checks: list
    :return: str
    """
    if False in checks and True in checks:
        return STATUS_ERROR
    elif True in checks:
        return STATUS_OK
    elif False in checks:
        return STATUS_NOT_SET

class BaseUtils:
    BASE_RIGHTS = 644
    BASE_RIGHTS_OCT = 0o644

    @staticmethod
    def exec_command(command, enable_stderr=False):
        #print(command)
        """
         Исполняет команду в BASH
         :param command: команда
         :param enable_stderr: bool
        """
        try:
            if enable_stderr:
                check_call(command, stdout=DEVNULL, shell=True)
            else:
                check_call(command, stdout=DEVNULL, stderr=DEVNULL, shell=True)
        except CalledProcessError as error:
            raise RbtConfigError('Команда \'%s\' завершилась c кодом возврата %d' % (command, error.returncode))

    @staticmethod
    def _check_output(command):
        """
        Исполняет команду в BASH с выводом результата исполнения команды :param command
        :return: str
        """
        try:
            result = check_output(command, shell=True, stderr=DEVNULL)
        except CalledProcessError as error:
            raise RbtConfigError('Команда \'%s\' завершилась с кодом возврата %d' % (command, error.returncode))
        return result.decode().rstrip('\n')

    @staticmethod
    def _check_call(command):
        """
        Возвращает True, если команда :param command выполнена успешно и False в противном случае
        :param command: str
        :return: bool
        """
        try:
            check_call(command, stdout=DEVNULL, stderr=DEVNULL, shell=True)
        except CalledProcessError:
            return False
        return True

    def check_is_user_in_group(self, user, group):
        """
         Проверяет, состоит ли пользователь user в группе group
         :param user: Имя пользователя (str)
         :param group: Имя группы (str)
        """
        return self._check_call('groups %s | grep %s' % (user, group))

    def chown(self, path, owner):
        self.exec_command('chown %s.%s %s' % (owner, owner, path))

    @staticmethod
    def rights(path):
        """
        Возвращает права на файл :param path
        :param path:
        :return: int
        """
        return int(oct(S_IMODE(os.stat(path).st_mode))[2:])

    @staticmethod
    def chmod(path, permissions=BASE_RIGHTS_OCT):
        return os.chmod(path, permissions)

    @staticmethod
    def owner(path):
        """
        Возвращает владельца файла :param path
        :param path: str
        :return: str
        """
        return getpwuid(os.stat(path).st_uid).pw_name

    @staticmethod
    def path_exists(path):
        return exists(path)

    @staticmethod
    def dirname(path):
        return dirname(path)

    @staticmethod
    def hostname():
        """
        Метод получения имени машины
        :return: str
        """
        try:
            result = BaseUtils._check_output('hostname -f')
        except RbtConfigError:
            raise RbtConfigError('Ошибка получения имени домена')
        return result

    @staticmethod
    def ip():
        """
        Метод получения IP-адреса машины
        :return:
        """
        try:
            result = BaseUtils._check_output('hostname -i').split(' ')[0:][0]
        except RbtConfigError:
            raise RbtConfigError('Ошибка получения IP-адреса')
        return result

    @staticmethod
    def subnet():
        """
        Метод получения подсети машины
        :return: str
        """
        try:
            ip = BaseUtils.ip()
        except RbtConfigError:
            raise RbtConfigError('Ошибка получения IP-адреса подсети')
        if ip == '127.0.0.1':
            return ip
        else:
            sub = ip.split('.')
            sub[3] = '0'
            return '.'.join(sub)

    @staticmethod
    def remove_file(path):
        if exists(path):
            os.remove(path)

    def check_if_package_exists(self, package_name):
        """
        Проверяет наличие пакета с именем :param package_name
        :param package_name: str
        :return: bool
        group(1) - имя пакета
        group(2) - дополнительная приставка к имени пакета
                   :amd64 или что нибудь еще
                   содержание приставки не обязательно(хз правильно ли это или нет)
        group(3) - статус пакета
        """
        text = self._check_output('dpkg --get-selections %s' % package_name)
        reg = r'^(%s)(:\S+)?\s+(\w+)' % package_name
        result = match(reg, text)
        if result:     
            status = result.group(3)
            if status and status == 'install':
                return True
        return False


class LocalUtils(BaseUtils):
    def __init__(self, username):
        self.__username = username

    def add_user_to_shadow_group(self):
        """
         Добавляет пользователя user в группу shadow
        """
        self.exec_command('usermod -aG shadow %s' % self.__username)

    def set_facl_on_parsec_db_to_user(self):
        """
         Установка прав на работу с PARSEC-DB пользователю user
        """
        db_list = ['macdb', 'capdb', 'micdb', 'auddb']
        for db in db_list:
            self.exec_command('setfacl -d -m u:%s:r /etc/parsec/%s' % (self.__username, db))
            self.exec_command('setfacl -R -m u:%s:r /etc/parsec/%s' % (self.__username, db))
            self.exec_command('setfacl -m u:%s:rx /etc/parsec/%s' % (self.__username, db))

    def check_facl_on_parsec_db(self):
        """
         Проверяет наличие прав на базы PARSEC
        """
        db_list = ['macdb', 'capdb', 'micdb', 'auddb']
        checks = set()
        for db in db_list:
            if not self._check_call('getfacl /etc/parsec/%s | grep %s' % (db, self.__username)):
                checks.add(False)
            else:
                checks.add(True)
        return all(checks)

    def del_facl_on_parsec_db_from_user(self):
        """
         Сброс установленных ранее прав пользователя user на PARSEC-DB
        """
        db_list = ['macdb', 'capdb', 'micdb', 'auddb']
        for db in db_list:
            self.exec_command('setfacl -x u:%s /etc/parsec/%s' % (self.__username, db))
            self.exec_command('setfacl -d -x u:%s /etc/parsec/%s' % (self.__username, db))

    def set_facl(self, rights, path):
        """
        Устанавливает права :param rights на путь :param path для текущего пользователя
        :param rights: права (str)
        :param path: путь (str)
        """
        self.exec_command('setfacl -m u:%s:%s %s' % (self.__username, rights, path))

    def check_facl(self, path, rights):
        """
        Проверяет наличие прав :param rights по пути :param path для текущего пользователя
        :param path: путь (str)
        :param rights: права (str)
        :return: bool
        """
        return self._check_call('getfacl %s | grep %s | grep %s' % (path, self.__username, rights))

    def del_facl(self, path):
        """
        Сбрасывает права :param rights по пути :path для текущего пользователя
        :param path: путь (str)
        """
        self.exec_command('setfacl -x u:%s %s' % (self.__username, path))
        self.exec_command('setfacl -d -x u:%s %s' % (self.__username, path))

    def del_user_from_shadow_group(self):
        """
         Удаляет пользователя user из группы shadow
        """
        self.exec_command('gpasswd -d %s shadow' % self.__username)

    def check_is_user_in_shadow_group(self):
        """
         Проверяет, состоит ли пользователь в групппе shadow?
        """
        return self.check_is_user_in_group(self.__username, 'shadow')


class AldUtils(BaseUtils):
    ALD_PASS_RIGHTS = 400
    ALD_PASS_RIGHTS_OCT = 0o400
    __PRINCIPAL_COMMAND_GETTING = 'klist -k %s | grep -v ^K | grep -v ^- | uniq'

    def __init__(self, service_owner, keytab_directory, principal_name,
                 keytab_file_name=None, auth_data=None, chown_keytab_owner=True):
        self.__principal = principal_name
        self.__service_owner = service_owner
        self.__ald_pass_file = '%s/.%s' % (expanduser('~'), 'ald-pass')
        if keytab_file_name:
            self.__krb_keytab = '%s/%s' % (keytab_directory, keytab_file_name)
        else:
            self.__krb_keytab = '%s/krb5.keytab' % keytab_directory
        self.__hostname = self.hostname()
        self.__chown_keytab_owner = chown_keytab_owner
        self.__make_ald_pass_if_not_exists(auth_data)

    def __del__(self):
        """
        Удаляем ~/.ald-pass файл перед разрушением объекта AldUtils
        """
        if exists(self.__ald_pass_file):
            os.remove(self.__ald_pass_file)

    def __get_principal_names(self):
        """
        Возвращает имя принципалов Kerberos из keytab
        :return: list of str or None
        """
        try:
            out = check_output(self.__PRINCIPAL_COMMAND_GETTING % self.__krb_keytab, shell=True, stderr=DEVNULL)
        except CalledProcessError:
            return None
        out = out.decode().rstrip('\n')
        if out:
            principals = []
            for line in out.split('\n'):
                try:
                    principal = [item for item in line.split(' ') if item not in [None, '', ' ']][1]
                    principal = [item for item in principal.split('@')][0]
                    principals.append(principal)
                except IndexError:
                    return None
            return principals
        return None

    def __make_ald_pass_if_not_exists(self, auth_data):
        """
        Создает в домашнем каталоге пользователя файл ~/.ald-pass с правами 400
        с данными для аутентификацией :param auth_data
        :param auth_data: str
        :return: None
        """
        if auth_data and not exists(self.__ald_pass_file):
            try:
                keytab = open(self.__ald_pass_file, mode='w')
                keytab.write(auth_data)
                keytab.close()
                BaseUtils.chmod(self.__ald_pass_file, AldUtils.ALD_PASS_RIGHTS_OCT)
            except OSError:
                raise RbtConfigError('Ошибка создания keytab для сервиса')

    @staticmethod
    def realm():
        """
        Получить имя realm сервера
        :return: str
        """
        hostname = BaseUtils.hostname()
        return '.'.join(hostname.split('.')[1:]).upper()

    def keytab(self):
        """
        Возвращает путь к keytab сервиса
        :return:
        """
        return self.__krb_keytab

    def check_if_ald_pass_file_exits(self):
        """
        Проверяет наличие парольного файла ALD и корректность установленных на него прав
        :return: bool
        """
        result = exists(self.__ald_pass_file)
        if result:
            result = result and self.rights(self.__ald_pass_file) == AldUtils.ALD_PASS_RIGHTS
        return result

    def add_service_to_domain(self):
        """
         Добавляет сервис в текущий домен.
        """
        self.exec_command('ald-admin service-add %s/%s --pass-file %s' % (self.__principal,
                                                                          self.__hostname,
                                                                          self.__ald_pass_file))

    def check_is_service_in_domain(self):
        """
        Проверяет наличие сервиса в домене
        :return: bool
        """
        return self._check_call('ald-admin service-get %s/%s --pass-file %s' % (self.__principal,
                                                                                self.__hostname,
                                                                                self.__ald_pass_file))

    def add_service_to_group(self, group):
        """
         Добавляет сервис в группу group
         :param group: Имя группы (str)
        """
        self.exec_command('ald-admin sgroup-svc-add %s/%s --sgroup=%s --pass-file %s' % (self.__principal,
                                                                                         self.__hostname,
                                                                                         group,
                                                                                         self.__ald_pass_file))

    def add_user_group_to_domain(self, group, gid):
        """
        Добавляет группу :param group в ALD с gid :param gid
        :param group: str
        :param gid: int
        :return: None
        """
        self.exec_command('ald-admin group-add %s --gid %d --pass-file %s -f ' % (group,
                                                                                  gid,
                                                                                  self.__ald_pass_file))

    def check_is_user_group_in_domain(self, group):
        """
        Проверяет наличие пользовательской группы :param group в домене
        :param group: str
        :return: bool
        """
        return self._check_call('ald-admin group-get %s --pass-file %s' % (group,
                                                                           self.__ald_pass_file))

    def del_user_group_from_domain(self, group):
        """
        Удаляет группу :param group из ALD
        :param group: str
        :return: None
        """
        self.exec_command('ald-admin group-rm %s --pass-file %s -f' % (group,
                                                                       self.__ald_pass_file))

    def create_keytab_for_service(self):
        """
         Создаёт keytab для сервиса.
        """
        self.exec_command('ald-client update-svc-keytab %s/%s --ktfile="%s" --pass-file %s'
                          % (self.__principal, self.__hostname, self.__krb_keytab, self.__ald_pass_file))
        if self.__chown_keytab_owner:
            self.chown(self.__krb_keytab, self.__service_owner)
        self.chmod(self.__krb_keytab)

    def check_is_ald_running(self):
        """
        Проверяет, запущен ли ALD
        :return:
        """
        return self._check_call('ald-client status --client && ald-client check-ldap-conn')

    def check_is_exists_keytab(self):
        """
        Проверяет наличие keytab сервиса корректность прав и наличия принциппа внутри него
        :return: bool
        """
        result = exists(self.__krb_keytab)
        if result:
            result = result and (self.rights(self.__krb_keytab) == BaseUtils.BASE_RIGHTS)
            if self.__chown_keytab_owner:
                result = result and (self.owner(self.__krb_keytab) == self.__service_owner)
            principals = self.__get_principal_names()
            if principals:
                expected_principal = '%s/%s' % (self.__principal, self.__hostname)
                result = result and (expected_principal in principals)
            else:
                return False
        return result

    def remove_keytab(self):
        """
        Удаляет принципал keytab'а из файла
        """
        if exists(self.__krb_keytab):
            self.exec_command('ald-client update-svc-keytab --remove %s/%s --ktfile="%s" --pass-file %s' % (
                self.__principal, self.__hostname, self.__krb_keytab, self.__ald_pass_file))

    def del_service_from_group(self, group):
        """
         Удаляет сервис из указанной группы.
         :param group: Имя группы (str)
        """
        self.exec_command('ald-admin sgroup-svc-rm %s/%s --sgroup=%s -f --pass-file %s'
                          % (self.__principal, self.__hostname, group, self.__ald_pass_file))

    def check_is_service_in_group(self, group):
        """
        Проверяет наличие сервиса в группе
        :param group: Имя группы (str)
        :return: bool
        """
        return self._check_call('ald-admin sgroup-get %s --pass-file %s | grep %s/%s' % (group,
                                                                                         self.__ald_pass_file,
                                                                                         self.__principal,
                                                                                         self.__hostname
                                                                                         ))

    def del_service_from_domain(self):
        """
         Удаляет service из домена domain.
        """

        self.exec_command('ald-admin service-rm %s/%s -f --pass-file %s' % (self.__principal,
                                                                            self.__hostname,
                                                                            self.__ald_pass_file))


def print_stderr(message, end='\n'):
    print(message, file=stderr, end=end)


def print_stdout(message):
    print(message)
    exit(0)


def exit_with_error(message):
    print(message, file=stderr)
    exit(EXIT_CODE_ERROR)
