#!/usr/bin/env python3
from __init__ import *

class Dovecot(BaseScript):
    def __init__(self, params):
        service_name = 'imap'
        service_config_dir = '/etc/dovecot'
        service_owner = 'dovecot'
        available_modes = [MODE_PAM, MODE_ALD]
        self.ald_utils = AldUtils(service_config_dir, service_owner, service_name)
        self.__krb_keytab = '/var/lib/dovecot/dovecot.keytab'
        self.__auth_conf = '%s/conf.d/10-auth.conf' % service_config_dir
        self.__master_conf = '%s/conf.d/10-master.conf' % service_config_dir
        self.__dovecot_conf = '%s/dovecot.conf' % service_config_dir
        self.__master_conf_data = ['unix_listener auth-client {',
                                   'mode = 0600',
                                   'user = Debian-exim',
                                   '}'
                                   ]
        self.__pam_script = '/etc/pam.d/dovecot'
        self.__pam_script_data = ['account required pam_tally.so']
        self.__ssl_conf = '%s/conf.d/10-ssl.conf' % service_config_dir
        self.__overridden_conf_path = '/etc/systemd/system/dovecot.service.d/astra_override.conf'
        params.update({
            INCOMPATIBLE_MODES: [(MODE_PAM, MODE_ALD)],
            KEYTAB_DIRECTORY: '/var/lib/dovecot',
            KEYTAB_FILE_NAME: 'dovecot.keytab',
            CHOWN_KEYTAB_OWNER: False,
            INIT_D_SERVICE_NAME: 'dovecot',
            REQUIRED_PACKAGES: {
                MODE_PAM: ['dovecot-imapd'],
                MODE_ALD: ['dovecot-imapd', 'dovecot-gssapi']
            },
        })
        super().__init__(service_name,
                         service_owner,
                         service_config_dir=service_config_dir,
                         available_modes=available_modes,
                         params=params
                         )

    def check_pam(self):
        checks = list()
        if self.local_utils.path_exists(self.__auth_conf):
            plain_auth_enabled = self.config_parser.config_file_contains_key_value(self.__auth_conf,
                                                                                   'auth_mechanisms',
                                                                                   'plain')
            if not plain_auth_enabled:
                checks.append(plain_auth_enabled)
            checks.append(self.config_parser.config_file_contains_key_value(self.__auth_conf,
                                                                            'disable_plaintext_auth', 'no'))
        else:
            checks.append(False)
        if self.local_utils.path_exists(self.__master_conf):
            checks.append(self.config_parser.config_file_contains_rows(self.__master_conf, self.__master_conf_data))
        else:
            checks.append(False)
        if self.local_utils.path_exists(self.__pam_script):
            checks.append(self.config_parser.config_file_contains_rows(self.__pam_script, self.__pam_script_data))
        else:
            checks.append(False)
        if self.local_utils.path_exists(self.__ssl_conf):
            checks.append(self.config_parser.config_file_contains_key_value(self.__ssl_conf, 'ssl', 'no'))
        else:
            checks.append(False)
        if self.local_utils.path_exists(self.__dovecot_conf):
            checks.append(self.config_parser.config_file_contains_rows(self.__dovecot_conf, ['listen = %s' % self.local_utils.ip()]))
        else:
            checks.append(False)
        return analyze_check(checks)

    def configure_pam(self):
        self.config_parser.set_or_replace_key_in_config(self.__auth_conf, 'auth_mechanisms', 'plain')
        self.config_parser.set_or_replace_key_in_config(self.__auth_conf, 'disable_plaintext_auth', 'no')
        self.config_parser.add_after(self.__master_conf, self.__master_conf_data, 'service auth {')
        self.config_parser.set_or_replace_key_in_config(self.__ssl_conf, 'ssl', 'no')
        self.config_parser.add_after(self.__pam_script, self.__pam_script_data)
        self.config_parser.set_or_replace_key_in_config(self.__dovecot_conf,
                                                        'listen',
                                                        self.local_utils.ip())
        self.exec_function_with_inverse_call(direct=self._create_overridden_conf,
                                             inverse=self._delete_overridden_conf)
        self._daemon_reload()
        self.exec_function_with_inverse_call(direct=self.add_to_autoloading,
                                             inverse=self.remove_from_autoloading)

    def reset_pam(self):
        self.config_parser.remove_rows_from_config(self.__auth_conf,
                                                   ['disable_plaintext_auth = no',
                                                    'auth_mechanisms = plain'
                                                   ])   
        self.config_parser.remove_folow_rows_from_config(self.__master_conf, self.__master_conf_data)
        self.config_parser.remove_rows_from_config(self.__ssl_conf,
                                                   ['ssl = no'])
        self.config_parser.remove_rows_from_config(self.__pam_script,
                                                   self.__pam_script_data)
        self.config_parser.set_or_replace_key_in_config(self.__dovecot_conf,
                                                        'listen', '*')
        self.exec_function_with_inverse_call(direct=self._delete_overridden_conf,
                                             inverse=self._create_overridden_conf)
        self._daemon_reload()
        self.exec_function_with_inverse_call(direct=self.remove_from_autoloading,
                                             inverse=self.add_to_autoloading)

    def configure_ald(self):
        self.config_parser.set_or_replace_key_in_config(self.__auth_conf, 'disable_plaintext_auth', 'yes')
        self.config_parser.set_or_replace_key_in_config(self.__auth_conf, 'auth_mechanisms', 'gssapi')
        self.config_parser.set_or_replace_key_in_config(self.__auth_conf, 'auth_gssapi_hostname',
                                                        self.ald_utils.hostname())
        self.config_parser.set_or_replace_key_in_config(self.__auth_conf, 'auth_krb5_keytab',
                                                        self.ald_utils.keytab())
        self.config_parser.add_after(self.__master_conf, self.__master_conf_data, 'service auth {')
        if not self.ald_utils.check_is_service_in_domain():
            self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_domain,
                                                 inverse=self.ald_utils.del_service_from_domain)
        if not self.ald_utils.check_is_service_in_group('mac'):
            self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_group,
                                                 inverse=self.ald_utils.del_service_from_group,
                                                 direct_args=['mac'])
        if not self.ald_utils.check_is_service_in_group('mail'):
            self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_group,
                                                 inverse=self.ald_utils.del_service_from_group,
                                                 direct_args=['mail'])
        self.exec_function_with_inverse_call(direct=self.ald_utils.create_keytab_for_service,
                                             inverse=self.ald_utils.remove_keytab)
        self.exec_function_with_inverse_call(direct=self.local_utils.set_facl,
                                             inverse=self.local_utils.del_facl,
                                             direct_args=['x', self.local_utils.dirname(self.ald_utils.keytab())])
        self.exec_function_with_inverse_call(direct=self.local_utils.set_facl,
                                             inverse=self.local_utils.del_facl,
                                             direct_args=['r', self.ald_utils.keytab()])
        self.config_parser.set_or_replace_key_in_config(self.__ssl_conf, 'ssl', 'no')
        self.exec_function_with_inverse_call(direct=self.add_to_autoloading,
                                             inverse=self.remove_from_autoloading)

    def check_ald(self):
        checks = list()
        checks.append(self.ald_utils.check_is_exists_keytab())
        if self.ald_utils.check_if_ald_pass_file_exits():
            checks.append(self.ald_utils.check_is_service_in_domain())
            checks.append(self.ald_utils.check_is_service_in_group('mac'))
            checks.append(self.ald_utils.check_is_service_in_group('mail'))
        if self.local_utils.path_exists(self.__auth_conf):
            checks.append(self.config_parser.config_file_contains_key_value(self.__auth_conf,
                                                                            'disable_plaintext_auth', 'yes'))
            checks.append(self.config_parser.config_file_contains_key_value(self.__auth_conf, 'auth_mechanisms', 'gssapi'))
            checks.append(self.config_parser.config_file_contains_key_value(self.__auth_conf, 'auth_gssapi_hostname',
                                                                            self.ald_utils.hostname()))
            checks.append(self.config_parser.config_file_contains_key_value(self.__auth_conf, 'auth_krb5_keytab',
                                                                            self.ald_utils.keytab()))            
        else:
            checks.append(False)
        if self.local_utils.path_exists(self.__ssl_conf):
            checks.append(self.config_parser.config_file_contains_key_value(self.__ssl_conf, 'ssl', 'no'))
        else:
            checks.append(False)
            
        return analyze_check(checks)

    def reset_ald(self):
        self.config_parser.remove_rows_from_config(self.__auth_conf, 'disable_plaintext_auth = yes')
        self.config_parser.remove_rows_from_config(self.__auth_conf, 'auth_mechanisms = gssapi')
        self.config_parser.remove_rows_from_config(self.__auth_conf, 'auth_gssapi_hostname = self.ald_utils.hostname()')        
        self.config_parser.remove_folow_rows_from_config(self.__master_conf, self.__master_conf_data)        
        
        self.exec_function_with_inverse_call(direct=self.ald_utils.remove_keytab,
                                             inverse=self.ald_utils.create_keytab_for_service)
        if self.ald_utils.check_is_service_in_group('mac'):
            self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_group,
                                                 inverse=self.ald_utils.add_service_to_group,
                                                 direct_args=['mac'])
        if self.ald_utils.check_is_service_in_group('mail'):
            self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_group,
                                                 inverse=self.ald_utils.add_service_to_group,
                                                 direct_args=['mail'])
        self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_domain,
                                             inverse=self.ald_utils.add_service_to_domain)
        self.config_parser.remove_rows_from_config(self.__ssl_conf,
                                                   ['ssl = no'])
        self.exec_function_with_inverse_call(direct=self.remove_from_autoloading,
                                             inverse=self.add_to_autoloading)

    def _create_overridden_conf(self) -> None:
        dirpath = os.path.dirname(self.__overridden_conf_path)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        with open(self.__overridden_conf_path, 'w') as file:
            file.write('''# DO NOT EDIT THIS FILE

[Unit]
Wants=network-online.target systemd-tmpfiles-setup.service
After=network-online.target systemd-tmpfiles-setup.service
''')

    def _delete_overridden_conf(self) -> None:
        if os.path.exists(self.__overridden_conf_path):
            os.remove(self.__overridden_conf_path)


class Exim(BaseScript):
    def __init__(self, params):
        service_name = 'smtp'
        service_config_dir = '/etc/exim4'
        service_owner = 'root'
        available_modes = [MODE_PAM, MODE_ALD]
        params.update({
            INCOMPATIBLE_MODES: [(MODE_PAM, MODE_ALD)],
            REQUIRED_PACKAGES: {
                MODE_PAM: ['exim4-daemon-heavy'],
                MODE_ALD: ['exim4-daemon-heavy']
            },
            KEYTAB_DIRECTORY: '/var/lib/dovecot',
            KEYTAB_FILE_NAME: 'dovecot.keytab',
            CHOWN_KEYTAB_OWNER: False,
            INIT_D_SERVICE_NAME: 'exim4'
        })
        self.__exim_conf_file = '%s/update-exim4.conf.conf' % service_config_dir
        self.__mailname_file = '/etc/mailname'
        self.__panic_log = '/var/log/exim4/paniclog'
        self.__config_check_rcpt_file = '%s/conf.d/acl/30_exim4-config_check_rcpt' % service_config_dir
        self.__config_files = {
            MODE_PAM: '%s/conf.d/auth/05_dovecot_login' % service_config_dir,
            MODE_ALD: '%s/conf.d/auth/33_exim4-dovecot-kerberos-ald' % service_config_dir
        }
        self.__configs_content = {
            MODE_PAM: ['dovecot_plain:',
                       'driver = dovecot',
                       'public_name = plain',
                       'server_socket = /var/run/dovecot/auth-client',
                       'server_set_id = $auth1'
                       ],
            MODE_ALD: ['dovecot_gssapi:',
                       'driver = dovecot',
                       'public_name = GSSAPI',
                       'server_socket = /var/run/dovecot/auth-client',
                       'server_set_id = $auth1'
                       ]
        }
        self.__config_check_rcpt_content = ['deny',
                                            'message = "Auth required"',
                                            'hosts = *:+relay_from_hosts',
                                            '!authenticated = *'
                                            ]
        self.__realm = AldUtils.realm().lower()
        if not self.__realm:
            raise RbtConfigError('Ошибка определения имени домена.')        
        self.__configs = {
            'dc_eximconfig_configtype': 'internet',
            'dc_other_hostnames': '%s' % self.__realm,
            'dc_local_interfaces': '%s' % BaseUtils.ip(),
            'dc_relay_domains': '',
            'dc_relay_nets': '',
            'dc_minimaldns': 'false',
            'dc_use_split_config': 'true',
            'dc_localdelivery': 'maildir_home'
        }
        super().__init__(service_name,
                         service_owner,
                         service_config_dir=service_config_dir,
                         available_modes=available_modes,
                         params=params
                         )

    def check_is_service_in_autoloading(self):
        """
         Проверяет наличие сервиса в автозагрузке
         :return bool
         """
        return self.local_utils._check_call('systemctl is-enabled %s' % self.init_d_service_name)

    def check_configuration(self):
        """
        Проверяет конфигурацию exim
        :return: bool
        """
        results = list()
        if self.local_utils.path_exists(self.__exim_conf_file):
            for key in self.__configs:
                results.append(self.config_parser.config_file_contains_key_value(self.__exim_conf_file,
                                                                                 key,
                                                                                 self.__configs[key],
                                                                                 need_escape_string=True))
        else:
            results.append(False)
        if self.local_utils.path_exists(self.__mailname_file):
            results.append(self.config_parser.config_file_contains_only_one_row(self.__mailname_file,
                                                                                self.__realm))
        else:
            results.append(False)
        if self.local_utils.path_exists(self.__config_check_rcpt_file):
            results.append(self.config_parser.config_file_contains_rows(self.__config_check_rcpt_file,
                                                                        self.__config_check_rcpt_content))
        else:
            results.append(False)
        return all(results)

    def set_configuration(self):
        """
        Выполняет общую настройку конфигурацию exim
        """
        for key in self.__configs:
            self.config_parser.set_or_replace_key_in_config(self.__exim_conf_file,
                                                            key,
                                                            self.__configs[key],
                                                            need_escape_strings=True,
                                                            remove_spaces=True)
        self.config_parser.insert_row_to_config_file(self.__mailname_file, self.__realm)
        self.config_parser.add_after(self.__config_check_rcpt_file, self.__config_check_rcpt_content, 'acl_check_rcpt:')

    def check_pam(self):
        checks = list()
        init_rd_check = self.check_is_service_in_autoloading()
        if not init_rd_check:
            checks.append(init_rd_check)        
        configuration_check = self.check_configuration()
        if not configuration_check:
            checks.append(configuration_check)
        if self.local_utils.path_exists(self.__config_files[MODE_PAM]):
            checks.append(self.config_parser.config_file_contains_rows(self.__config_files[MODE_PAM],
                                                                       self.__configs_content[MODE_PAM]))
        else:
            checks.append(False)
        return analyze_check(checks)

    def configure_pam(self):
        self.exec_function_with_inverse_call(direct=self.add_to_autoloading,
                                             inverse=self.remove_from_autoloading)        
        self.local_utils.remove_file(self.__panic_log)
        self.set_configuration()
        try:
            self.local_utils.exec_command('rm -f /var/mail/*')
        except RbtConfigError:
            pass
        self.config_parser.write_file_with_content(self.__config_files[MODE_PAM], self.__configs_content[MODE_PAM])

    def reset_pam(self):
        self.local_utils.remove_file(self.__config_files[MODE_PAM])
        self.exec_function_with_inverse_call(direct=self.remove_from_autoloading,
                                             inverse=self.add_to_autoloading)        

    def configure_ald(self):
        self.local_utils.remove_file(self.__panic_log)
        self.set_configuration()
        self.config_parser.write_file_with_content(self.__config_files[MODE_ALD], self.__configs_content[MODE_ALD])
        self.exec_function_with_inverse_call(direct=self.add_to_autoloading,
                                             inverse=self.remove_from_autoloading)   
        if not self.ald_utils.check_is_service_in_domain():
            self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_domain,
                                                 inverse=self.ald_utils.del_service_from_domain)
        if not self.ald_utils.check_is_service_in_group('mac'):
            self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_group,
                                                 inverse=self.ald_utils.del_service_from_group,
                                                 direct_args=['mac'])
        if not self.ald_utils.check_is_service_in_group('mail'):
            self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_group,
                                                 inverse=self.ald_utils.del_service_from_group,
                                                 direct_args=['mail'])
        self.exec_function_with_inverse_call(direct=self.ald_utils.create_keytab_for_service,
                                             inverse=self.ald_utils.remove_keytab)

    def check_ald(self):
        checks = list()
        init_rd_check = self.check_is_service_in_autoloading()
        if not init_rd_check:
            checks.append(init_rd_check)        
        configuration_check = self.check_configuration()
        if not configuration_check:
            checks.append(configuration_check)
        checks.append(self.ald_utils.check_is_exists_keytab())
        if self.ald_utils.check_if_ald_pass_file_exits():
            checks.append(self.ald_utils.check_is_service_in_domain())
            checks.append(self.ald_utils.check_is_service_in_group('mac'))
            checks.append(self.ald_utils.check_is_service_in_group('mail'))
        if self.local_utils.path_exists(self.__config_files[MODE_ALD]):
            checks.append(self.config_parser.config_file_contains_rows(self.__config_files[MODE_ALD],
                                                                       self.__configs_content[MODE_ALD]))
        else:
            checks.append(False)
        return analyze_check(checks)

    def reset_ald(self):
        self.local_utils.remove_file(self.__config_files[MODE_ALD])
        self.exec_function_with_inverse_call(direct=self.remove_from_autoloading,
                                             inverse=self.add_to_autoloading)        
        self.exec_function_with_inverse_call(direct=self.ald_utils.remove_keytab,
                                             inverse=self.ald_utils.create_keytab_for_service)
        if self.ald_utils.check_is_service_in_group('mac'):
            self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_group,
                                                 inverse=self.ald_utils.add_service_to_group,
                                                 direct_args=['mac'])
        if self.ald_utils.check_is_service_in_group('mail'):
            self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_group,
                                                 inverse=self.ald_utils.add_service_to_group,
                                                 direct_args=['mail'])
        self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_domain,
                                             inverse=self.ald_utils.add_service_to_domain)


if __name__ == '__main__':
    run([Dovecot, Exim], param_required=False)
