#!/usr/bin/env python3
# Проверка на наличие установленных модулей
from __init__ import *

class Postgresql(BaseScript):
    def __init__(self, params):
        service_name = 'postgres'
        service_owner = 'postgres'
        available_modes = [MODE_PAM, MODE_ALD]
        try:
            self.__cluster = params['cluster']
            self.__version = params['version']
        except KeyError:
            raise RbtConfigError('Ошибка получения параметров')
        self.__service_config_dir = '/etc/postgresql/%s/%s' % (self.__version, self.__cluster)
        self.__postgresql_conf = '%s/postgresql.conf' % self.__service_config_dir
        params.update({
            # PostgreSQL может одновременно работать в ЕПП и PAM режимах,
            # Но мы обеспечиваем в качестве базовой настройки сервиса PostgreSQL только *один*
            INCOMPATIBLE_MODES: [(MODE_ALD, MODE_PAM)],
            KEYTAB_DIRECTORY: '/etc/postgresql-common'
        })
        super().__init__(service_name,
                         service_owner,
                         service_config_dir=self.__service_config_dir,
                         available_modes=available_modes,
                         params=params
                         )

    def set_auth(self, auth_type):
        """
        Устанавливает аутентификацию в pg_hba.conf кластера по методу :param auth_type
        :param auth_type: pam или gss (str)
        :return: None
        """
        if auth_type not in ['pam', 'gss']:
            raise RbtConfigError('Некоректный метод аутентификации: \'%s\'' % auth_type)
        hba_file = '%s/pg_hba.conf' % self.__service_config_dir
        if not exists(hba_file):
            raise RbtConfigError('Конфигурационный файл \'%s\' не найден' % hba_file)
        # Добавялем файл в бекап
        self.config_parser.append_file_to_backup(hba_file)
        lines = open(hba_file).readlines()
        expr = r'^(?!#)host|hostssl|hostnossl'
        results = []
        row_found = False
        for line in lines:
            if match(expr, line):
                # Ищем первую строку подключения,
                # Все остальные мы комментируем
                if not row_found:
                    attributes = [param for param in line.split(' ') if not param == '']
                    # Мы ожидаем для подключения host* 5 параметров:
                    # тип (host|hostssl|hostnossl, database, user, subnet, auth_type)
                    if len(attributes) < 5:
                        RbtConfigError('Ошибка разбора параметров конфигурации в строке %s файла %s' % (line, hba_file))
                    else:
                        # Пока реализована поддержка только адресов IPv4
                        if not attributes[3] == '::1/128':
                            attributes[0] = 'host'
                            attributes[1] = 'all'
                            attributes[2] = 'all'
                            attributes[3] = '%s/32' % self.local_utils.subnet()
                            attributes[4] = auth_type
                            attributes = attributes[:5]                            
                            if (auth_type == 'gss'):
                                attributes.append('include_realm=0')
                    line = '        '.join(attributes) + '\n'
                    row_found = True
                else:
                    line = '# %s' % line
            results.append(line)
        open(hba_file, mode='w').writelines(results)

    def check_auth(self, auth_type):
        """
        Выполняет проверку на тип аутентификации в pg_hba.conf
        :param auth_type: pam или gss (str)
        :return: bool
        """
        if auth_type not in ['pam', 'gss']:
            raise RbtConfigError('Некоректный метод аутентификации: \'%s\'' % auth_type)
        hba_file = '%s/pg_hba.conf' % self.__service_config_dir
        if not exists(hba_file):
            raise RbtConfigError('Конфигурационный файл \'%s\' не найден' % hba_file)
        lines = open(hba_file).readlines()
        expr = r'^(?!#)host|hostssl|hostnossl'
        for line in lines:
            if match(expr, line):
                attributes = [param for param in line.split(' ') if not param == '']
                if len(attributes) < 5:
                    RbtConfigError('Ошибка разбора параметров конфигурации в строке %s файла %s' % (line, hba_file))
                else:
                    # Пока реализована поддержка только адресов IPv4
                    if not attributes[3] == '::1/128':
                        return auth_type == attributes[4].rstrip('\n')
        return False

    def restart_service(self, raise_exception_if_error=True):
        """
        Переопределяем функцию перезапуска сервиса,
        т.к. для перезагрузки сервера PostgreSQL мы используем pg_ctlcluster
        """
        status = self.local_utils._check_call('pg_ctlcluster %s %s restart' % (self.__version, self.__cluster))
        if raise_exception_if_error and not status:
            raise RbtConfigError('Ошибка перезапуска кластера PostgreSQL')

    def check_pam(self):
        checks = [self.local_utils.check_facl_on_parsec_db(),
                  self.local_utils.check_is_user_in_shadow_group(),
                  self.check_auth('pam')
                  ]
        listen_addresses_check = self.config_parser.config_file_contains_key_value(self.__postgresql_conf,
                                                                                   'listen_addresses',
                                                                                   self.local_utils.ip(),
                                                                                   need_escape_string=True)
        ac_ignore_socket_maclabel_check = self.config_parser.config_file_contains_key_value(self.__postgresql_conf,
                                                                                            'ac_ignore_socket_maclabel',
                                                                                            'false',
                                                                                            need_escape_string=False)
        if not listen_addresses_check:
            checks.append(listen_addresses_check)
        if not ac_ignore_socket_maclabel_check:
            checks.append(ac_ignore_socket_maclabel_check)
        return analyze_check(checks)

    def configure_pam(self):
        self.exec_function_with_inverse_call(direct=self.local_utils.set_facl_on_parsec_db_to_user,
                                             inverse=self.local_utils.del_facl_on_parsec_db_from_user)
        self.exec_function_with_inverse_call(direct=self.local_utils.add_user_to_shadow_group,
                                             inverse=self.local_utils.del_user_from_shadow_group)
        self.config_parser.set_or_replace_key_in_config(self.__postgresql_conf,
                                                        'listen_addresses',
                                                        self.local_utils.ip(),
                                                        need_escape_strings=True)
        self.config_parser.set_or_replace_key_in_config(self.__postgresql_conf,
                                                        'ac_ignore_socket_maclabel',
                                                        'false',
                                                        need_escape_strings=False)
        self.set_auth('pam')

    def reset_pam(self):
        self.exec_function_with_inverse_call(direct=self.local_utils.del_facl_on_parsec_db_from_user,
                                             inverse=self.local_utils.set_facl_on_parsec_db_to_user)
        self.exec_function_with_inverse_call(direct=self.local_utils.del_user_from_shadow_group,
                                             inverse=self.local_utils.add_user_to_shadow_group)

    def configure_ald(self):
        if not self.ald_utils.check_is_exists_keytab():
            self.exec_function_with_inverse_call(direct=self.ald_utils.create_keytab_for_service,
                                                 inverse=self.ald_utils.create_keytab_for_service)
        if not self.ald_utils.check_is_service_in_domain():    
            self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_domain,
                                                 inverse=self.ald_utils.del_service_from_domain)
        if not self.ald_utils.check_is_service_in_group('mac'):
            self.exec_function_with_inverse_call(direct=self.ald_utils.add_service_to_group,
                                                 inverse=self.ald_utils.del_service_from_group,
                                                 direct_args=['mac'])
        self.config_parser.set_or_replace_key_in_config(self.__postgresql_conf,
                                                        'listen_addresses',
                                                        self.local_utils.ip(),
                                                        need_escape_strings=True)
        self.config_parser.set_or_replace_key_in_config(self.__postgresql_conf,
                                                        'ac_ignore_socket_maclabel',
                                                        'false',
                                                        need_escape_strings=False)
        self.set_auth('gss')

    def check_ald(self):
        checks = list()
        checks.append(self.ald_utils.check_is_exists_keytab())
        checks.append(self.check_auth('gss'))
        if self.ald_utils.check_if_ald_pass_file_exits():
            checks.append(self.ald_utils.check_is_service_in_domain())
            checks.append(self.ald_utils.check_is_service_in_group('mac'))
        listen_addresses_check = self.config_parser.config_file_contains_key_value(self.__postgresql_conf,
                                                                                   'listen_addresses',
                                                                                   self.local_utils.ip(),
                                                                                   need_escape_string=True)
        ac_ignore_socket_maclabel_check = self.config_parser.config_file_contains_key_value(self.__postgresql_conf,
                                                                                            'ac_ignore_socket_maclabel',
                                                                                            'false',
                                                                                            need_escape_string=False)
        if not listen_addresses_check:
            checks.append(listen_addresses_check)
        if not ac_ignore_socket_maclabel_check:
            checks.append(ac_ignore_socket_maclabel_check)
        return analyze_check(checks)

    def reset_ald(self):
        self.exec_function_with_inverse_call(direct=self.ald_utils.remove_keytab,
                                             inverse=self.ald_utils.create_keytab_for_service)
        self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_group,
                                             inverse=self.ald_utils.add_service_to_group,
                                             direct_args=['mac'])
        self.exec_function_with_inverse_call(direct=self.ald_utils.del_service_from_domain,
                                             inverse=self.ald_utils.add_service_to_domain)


if __name__ == '__main__':
    run(Postgresql, param_required=True)
